# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_persistance.ipynb.

# %% auto 0
__all__ = ['EasyDict', 'Logger', 'set_cache_dir', 'make_cache_dir_path', 'format_time', 'ask_yes_no', 'tuple_product',
           'get_dtype_and_ctype', 'is_pickleable', 'get_module_from_obj_name', 'get_obj_from_module', 'get_obj_by_name',
           'call_func_by_name', 'construct_class_by_name', 'get_module_dir_by_obj_name', 'is_top_level_function',
           'get_top_level_function_name', 'list_dir_recursively_with_ignore', 'copy_files_and_create_dirs', 'is_url',
           'open_url', 'persistent_class', 'is_persistent', 'import_hook']

# %% ../nbs/03_persistance.ipynb 2
import ctypes
import fnmatch
import importlib
import inspect
import numpy as np
import os
import shutil
import sys
import types
import io
import pickle
import re
import requests
import html
import hashlib
import glob
import tempfile
import urllib
import urllib.request
import uuid

from distutils.util import strtobool
from typing import Any, List, Tuple, Union


# Util classes
# ------------------------------------------------------------------------------------------


class EasyDict(dict):
    """Convenience class that behaves like a dict but allows access with the attribute syntax."""

    def __getattr__(self, name: str) -> Any:
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name: str, value: Any) -> None:
        self[name] = value

    def __delattr__(self, name: str) -> None:
        del self[name]


class Logger(object):
    """Redirect stderr to stdout, optionally print stdout to a file, and optionally force flushing on both stdout and the file."""

    def __init__(self, file_name: str = None, file_mode: str = "w", should_flush: bool = True):
        self.file = None

        if file_name is not None:
            self.file = open(file_name, file_mode)

        self.should_flush = should_flush
        self.stdout = sys.stdout
        self.stderr = sys.stderr

        sys.stdout = self
        sys.stderr = self

    def __enter__(self) -> "Logger":
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.close()

    def write(self, text: Union[str, bytes]) -> None:
        """Write text to stdout (and a file) and optionally flush."""
        if isinstance(text, bytes):
            text = text.decode()
        if len(text) == 0: # workaround for a bug in VSCode debugger: sys.stdout.write(''); sys.stdout.flush() => crash
            return

        if self.file is not None:
            self.file.write(text)

        self.stdout.write(text)

        if self.should_flush:
            self.flush()

    def flush(self) -> None:
        """Flush written text to both stdout and a file, if open."""
        if self.file is not None:
            self.file.flush()

        self.stdout.flush()

    def close(self) -> None:
        """Flush, close possible files, and remove stdout/stderr mirroring."""
        self.flush()

        # if using multiple loggers, prevent closing in wrong order
        if sys.stdout is self:
            sys.stdout = self.stdout
        if sys.stderr is self:
            sys.stderr = self.stderr

        if self.file is not None:
            self.file.close()
            self.file = None


# Cache directories
# ------------------------------------------------------------------------------------------

_dnnlib_cache_dir = None

def set_cache_dir(path: str) -> None:
    global _dnnlib_cache_dir
    _dnnlib_cache_dir = path

def make_cache_dir_path(*paths: str) -> str:
    if _dnnlib_cache_dir is not None:
        return os.path.join(_dnnlib_cache_dir, *paths)
    if 'DNNLIB_CACHE_DIR' in os.environ:
        return os.path.join(os.environ['DNNLIB_CACHE_DIR'], *paths)
    if 'HOME' in os.environ:
        return os.path.join(os.environ['HOME'], '.cache', 'dnnlib', *paths)
    if 'USERPROFILE' in os.environ:
        return os.path.join(os.environ['USERPROFILE'], '.cache', 'dnnlib', *paths)
    return os.path.join(tempfile.gettempdir(), '.cache', 'dnnlib', *paths)

# Small util functions
# ------------------------------------------------------------------------------------------


def format_time(seconds: Union[int, float]) -> str:
    """Convert the seconds to human readable string with days, hours, minutes and seconds."""
    s = int(np.rint(seconds))

    if s < 60:
        return "{0}s".format(s)
    elif s < 60 * 60:
        return "{0}m {1:02}s".format(s // 60, s % 60)
    elif s < 24 * 60 * 60:
        return "{0}h {1:02}m {2:02}s".format(s // (60 * 60), (s // 60) % 60, s % 60)
    else:
        return "{0}d {1:02}h {2:02}m".format(s // (24 * 60 * 60), (s // (60 * 60)) % 24, (s // 60) % 60)


def ask_yes_no(question: str) -> bool:
    """Ask the user the question until the user inputs a valid answer."""
    while True:
        try:
            print("{0} [y/n]".format(question))
            return strtobool(input().lower())
        except ValueError:
            pass


def tuple_product(t: Tuple) -> Any:
    """Calculate the product of the tuple elements."""
    result = 1

    for v in t:
        result *= v

    return result


_str_to_ctype = {
    "uint8": ctypes.c_ubyte,
    "uint16": ctypes.c_uint16,
    "uint32": ctypes.c_uint32,
    "uint64": ctypes.c_uint64,
    "int8": ctypes.c_byte,
    "int16": ctypes.c_int16,
    "int32": ctypes.c_int32,
    "int64": ctypes.c_int64,
    "float32": ctypes.c_float,
    "float64": ctypes.c_double
}


def get_dtype_and_ctype(type_obj: Any) -> Tuple[np.dtype, Any]:
    """Given a type name string (or an object having a __name__ attribute), return matching Numpy and ctypes types that have the same size in bytes."""
    type_str = None

    if isinstance(type_obj, str):
        type_str = type_obj
    elif hasattr(type_obj, "__name__"):
        type_str = type_obj.__name__
    elif hasattr(type_obj, "name"):
        type_str = type_obj.name
    else:
        raise RuntimeError("Cannot infer type name from input")

    assert type_str in _str_to_ctype.keys()

    my_dtype = np.dtype(type_str)
    my_ctype = _str_to_ctype[type_str]

    assert my_dtype.itemsize == ctypes.sizeof(my_ctype)

    return my_dtype, my_ctype


def is_pickleable(obj: Any) -> bool:
    try:
        with io.BytesIO() as stream:
            pickle.dump(obj, stream)
        return True
    except:
        return False


# Functionality to import modules/objects by name, and call functions by name
# ------------------------------------------------------------------------------------------

def get_module_from_obj_name(obj_name: str) -> Tuple[types.ModuleType, str]:
    """Searches for the underlying module behind the name to some python object.
    Returns the module and the object name (original name with module part removed)."""

    # allow convenience shorthands, substitute them by full names
    obj_name = re.sub("^np.", "numpy.", obj_name)
    obj_name = re.sub("^tf.", "tensorflow.", obj_name)

    # list alternatives for (module_name, local_obj_name)
    parts = obj_name.split(".")
    name_pairs = [(".".join(parts[:i]), ".".join(parts[i:])) for i in range(len(parts), 0, -1)]

    # try each alternative in turn
    for module_name, local_obj_name in name_pairs:
        try:
            module = importlib.import_module(module_name) # may raise ImportError
            get_obj_from_module(module, local_obj_name) # may raise AttributeError
            return module, local_obj_name
        except:
            pass

    # maybe some of the modules themselves contain errors?
    for module_name, _local_obj_name in name_pairs:
        try:
            importlib.import_module(module_name) # may raise ImportError
        except ImportError:
            if not str(sys.exc_info()[1]).startswith("No module named '" + module_name + "'"):
                raise

    # maybe the requested attribute is missing?
    for module_name, local_obj_name in name_pairs:
        try:
            module = importlib.import_module(module_name) # may raise ImportError
            get_obj_from_module(module, local_obj_name) # may raise AttributeError
        except ImportError:
            pass

    # we are out of luck, but we have no idea why
    raise ImportError(obj_name)


def get_obj_from_module(module: types.ModuleType, obj_name: str) -> Any:
    """Traverses the object name and returns the last (rightmost) python object."""
    if obj_name == '':
        return module
    obj = module
    for part in obj_name.split("."):
        obj = getattr(obj, part)
    return obj


def get_obj_by_name(name: str) -> Any:
    """Finds the python object with the given name."""
    module, obj_name = get_module_from_obj_name(name)
    return get_obj_from_module(module, obj_name)


def call_func_by_name(*args, func_name: str = None, **kwargs) -> Any:
    """Finds the python object with the given name and calls it as a function."""
    assert func_name is not None
    func_obj = get_obj_by_name(func_name)
    assert callable(func_obj)
    return func_obj(*args, **kwargs)


def construct_class_by_name(*args, class_name: str = None, **kwargs) -> Any:
    """Finds the python class with the given name and constructs it with the given arguments."""
    return call_func_by_name(*args, func_name=class_name, **kwargs)


def get_module_dir_by_obj_name(obj_name: str) -> str:
    """Get the directory path of the module containing the given object name."""
    module, _ = get_module_from_obj_name(obj_name)
    return os.path.dirname(inspect.getfile(module))


def is_top_level_function(obj: Any) -> bool:
    """Determine whether the given object is a top-level function, i.e., defined at module scope using 'def'."""
    return callable(obj) and obj.__name__ in sys.modules[obj.__module__].__dict__


def get_top_level_function_name(obj: Any) -> str:
    """Return the fully-qualified name of a top-level function."""
    assert is_top_level_function(obj)
    module = obj.__module__
    if module == '__main__':
        module = os.path.splitext(os.path.basename(sys.modules[module].__file__))[0]
    return module + "." + obj.__name__


# File system helpers
# ------------------------------------------------------------------------------------------

def list_dir_recursively_with_ignore(dir_path: str, ignores: List[str] = None, add_base_to_relative: bool = False) -> List[Tuple[str, str]]:
    """List all files recursively in a given directory while ignoring given file and directory names.
    Returns list of tuples containing both absolute and relative paths."""
    assert os.path.isdir(dir_path)
    base_name = os.path.basename(os.path.normpath(dir_path))

    if ignores is None:
        ignores = []

    result = []

    for root, dirs, files in os.walk(dir_path, topdown=True):
        for ignore_ in ignores:
            dirs_to_remove = [d for d in dirs if fnmatch.fnmatch(d, ignore_)]

            # dirs need to be edited in-place
            for d in dirs_to_remove:
                dirs.remove(d)

            files = [f for f in files if not fnmatch.fnmatch(f, ignore_)]

        absolute_paths = [os.path.join(root, f) for f in files]
        relative_paths = [os.path.relpath(p, dir_path) for p in absolute_paths]

        if add_base_to_relative:
            relative_paths = [os.path.join(base_name, p) for p in relative_paths]

        assert len(absolute_paths) == len(relative_paths)
        result += zip(absolute_paths, relative_paths)

    return result


def copy_files_and_create_dirs(files: List[Tuple[str, str]]) -> None:
    """Takes in a list of tuples of (src, dst) paths and copies files.
    Will create all necessary directories."""
    for file in files:
        target_dir_name = os.path.dirname(file[1])

        # will create all intermediate-level directories
        if not os.path.exists(target_dir_name):
            os.makedirs(target_dir_name)

        shutil.copyfile(file[0], file[1])


# URL helpers
# ------------------------------------------------------------------------------------------

def is_url(obj: Any, allow_file_urls: bool = False) -> bool:
    """Determine whether the given object is a valid URL string."""
    if not isinstance(obj, str) or not "://" in obj:
        return False
    if allow_file_urls and obj.startswith('file://'):
        return True
    try:
        res = requests.compat.urlparse(obj)
        if not res.scheme or not res.netloc or not "." in res.netloc:
            return False
        res = requests.compat.urlparse(requests.compat.urljoin(obj, "/"))
        if not res.scheme or not res.netloc or not "." in res.netloc:
            return False
    except:
        return False
    return True


def open_url(url: str, cache_dir: str = None, num_attempts: int = 10, verbose: bool = True, return_filename: bool = False, cache: bool = True) -> Any:
    """Download the given URL and return a binary-mode file object to access the data."""
    assert num_attempts >= 1
    assert not (return_filename and (not cache))

    # Doesn't look like an URL scheme so interpret it as a local filename.
    if not re.match('^[a-z]+://', url):
        return url if return_filename else open(url, "rb")

    # Handle file URLs.  This code handles unusual file:// patterns that
    # arise on Windows:
    #
    # file:///c:/foo.txt
    #
    # which would translate to a local '/c:/foo.txt' filename that's
    # invalid.  Drop the forward slash for such pathnames.
    #
    # If you touch this code path, you should test it on both Linux and
    # Windows.
    #
    # Some internet resources suggest using urllib.request.url2pathname() but
    # but that converts forward slashes to backslashes and this causes
    # its own set of problems.
    if url.startswith('file://'):
        filename = urllib.parse.urlparse(url).path
        if re.match(r'^/[a-zA-Z]:', filename):
            filename = filename[1:]
        return filename if return_filename else open(filename, "rb")

    assert is_url(url)

    # Lookup from cache.
    if cache_dir is None:
        cache_dir = make_cache_dir_path('downloads')

    url_md5 = hashlib.md5(url.encode("utf-8")).hexdigest()
    if cache:
        cache_files = glob.glob(os.path.join(cache_dir, url_md5 + "_*"))
        if len(cache_files) == 1:
            filename = cache_files[0]
            return filename if return_filename else open(filename, "rb")

    # Download.
    url_name = None
    url_data = None
    with requests.Session() as session:
        if verbose:
            print("Downloading %s ..." % url, end="", flush=True)
        for attempts_left in reversed(range(num_attempts)):
            try:
                with session.get(url) as res:
                    res.raise_for_status()
                    if len(res.content) == 0:
                        raise IOError("No data received")

                    if len(res.content) < 8192:
                        content_str = res.content.decode("utf-8")
                        if "download_warning" in res.headers.get("Set-Cookie", ""):
                            links = [html.unescape(link) for link in content_str.split('"') if "export=download" in link]
                            if len(links) == 1:
                                url = requests.compat.urljoin(url, links[0])
                                raise IOError("Google Drive virus checker nag")
                        if "Google Drive - Quota exceeded" in content_str:
                            raise IOError("Google Drive download quota exceeded -- please try again later")

                    match = re.search(r'filename="([^"]*)"', res.headers.get("Content-Disposition", ""))
                    url_name = match[1] if match else url
                    url_data = res.content
                    if verbose:
                        print(" done")
                    break
            except KeyboardInterrupt:
                raise
            except:
                if not attempts_left:
                    if verbose:
                        print(" failed")
                    raise
                if verbose:
                    print(".", end="", flush=True)

    # Save to cache.
    if cache:
        safe_name = re.sub(r"[^0-9a-zA-Z-._]", "_", url_name)
        cache_file = os.path.join(cache_dir, url_md5 + "_" + safe_name)
        temp_file = os.path.join(cache_dir, "tmp_" + uuid.uuid4().hex + "_" + url_md5 + "_" + safe_name)
        os.makedirs(cache_dir, exist_ok=True)
        with open(temp_file, "wb") as f:
            f.write(url_data)
        os.replace(temp_file, cache_file) # atomic
        if return_filename:
            return cache_file

    # Return data as file object.
    assert not return_filename
    return io.BytesIO(url_data)


# %% ../nbs/03_persistance.ipynb 3
# Copyright (c) 2021, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

"""Facilities for pickling Python code alongside other data.

The pickled code is automatically imported into a separate Python module
during unpickling. This way, any previously exported pickles will remain
usable even if the original code is no longer available, or if the current
version of the code is not consistent with what was originally pickled."""

import sys
import pickle
import io
import inspect
import copy
import uuid
import types

#----------------------------------------------------------------------------

_version            = 6         # internal version number
_decorators         = set()     # {decorator_class, ...}
_import_hooks       = []        # [hook_function, ...]
_module_to_src_dict = dict()    # {module: src, ...}
_src_to_module_dict = dict()    # {src: module, ...}

#----------------------------------------------------------------------------

def persistent_class(orig_class):
    r"""Class decorator that extends a given class to save its source code
    when pickled.

    Example:

        from torch_utils import persistence

        @persistence.persistent_class
        class MyNetwork(torch.nn.Module):
            def __init__(self, num_inputs, num_outputs):
                super().__init__()
                self.fc = MyLayer(num_inputs, num_outputs)
                ...

        @persistence.persistent_class
        class MyLayer(torch.nn.Module):
            ...

    When pickled, any instance of `MyNetwork` and `MyLayer` will save its
    source code alongside other internal state (e.g., parameters, buffers,
    and submodules). This way, any previously exported pickle will remain
    usable even if the class definitions have been modified or are no
    longer available.

    The decorator saves the source code of the entire Python module
    containing the decorated class. It does *not* save the source code of
    any imported modules. Thus, the imported modules must be available
    during unpickling, also including `torch_utils.persistence` itself.

    It is ok to call functions defined in the same module from the
    decorated class. However, if the decorated class depends on other
    classes defined in the same module, they must be decorated as well.
    This is illustrated in the above example in the case of `MyLayer`.

    It is also possible to employ the decorator just-in-time before
    calling the constructor. For example:

        cls = MyLayer
        if want_to_make_it_persistent:
            cls = persistence.persistent_class(cls)
        layer = cls(num_inputs, num_outputs)

    As an additional feature, the decorator also keeps track of the
    arguments that were used to construct each instance of the decorated
    class. The arguments can be queried via `obj.init_args` and
    `obj.init_kwargs`, and they are automatically pickled alongside other
    object state. A typical use case is to first unpickle a previous
    instance of a persistent class, and then upgrade it to use the latest
    version of the source code:

        with open('old_pickle.pkl', 'rb') as f:
            old_net = pickle.load(f)
        new_net = MyNetwork(*old_obj.init_args, **old_obj.init_kwargs)
        misc.copy_params_and_buffers(old_net, new_net, require_all=True)
    """
    assert isinstance(orig_class, type)
    if is_persistent(orig_class):
        return orig_class

    assert orig_class.__module__ in sys.modules
    orig_module = sys.modules[orig_class.__module__]
    orig_module_src = _module_to_src(orig_module)

    class Decorator(orig_class):
        _orig_module_src = orig_module_src
        _orig_class_name = orig_class.__name__

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._init_args = copy.deepcopy(args)
            self._init_kwargs = copy.deepcopy(kwargs)
            assert orig_class.__name__ in orig_module.__dict__
            _check_pickleable(self.__reduce__())

        @property
        def init_args(self):
            return copy.deepcopy(self._init_args)

        @property
        def init_kwargs(self):
            return EasyDict(copy.deepcopy(self._init_kwargs))

        def __reduce__(self):
            fields = list(super().__reduce__())
            fields += [None] * max(3 - len(fields), 0)
            if fields[0] is not _reconstruct_persistent_obj:
                meta = dict(type='class', version=_version, module_src=self._orig_module_src, class_name=self._orig_class_name, state=fields[2])
                fields[0] = _reconstruct_persistent_obj # reconstruct func
                fields[1] = (meta,) # reconstruct args
                fields[2] = None # state dict
            return tuple(fields)

    Decorator.__name__ = orig_class.__name__
    _decorators.add(Decorator)
    return Decorator

#----------------------------------------------------------------------------

def is_persistent(obj):
    r"""Test whether the given object or class is persistent, i.e.,
    whether it will save its source code when pickled.
    """
    try:
        if obj in _decorators:
            return True
    except TypeError:
        pass
    return type(obj) in _decorators # pylint: disable=unidiomatic-typecheck

#----------------------------------------------------------------------------

def import_hook(hook):
    r"""Register an import hook that is called whenever a persistent object
    is being unpickled. A typical use case is to patch the pickled source
    code to avoid errors and inconsistencies when the API of some imported
    module has changed.

    The hook should have the following signature:

        hook(meta) -> modified meta

    `meta` is an instance of `EasyDict` with the following fields:

        type:       Type of the persistent object, e.g. `'class'`.
        version:    Internal version number of `torch_utils.persistence`.
        module_src  Original source code of the Python module.
        class_name: Class name in the original Python module.
        state:      Internal state of the object.

    Example:

        @persistence.import_hook
        def wreck_my_network(meta):
            if meta.class_name == 'MyNetwork':
                print('MyNetwork is being imported. I will wreck it!')
                meta.module_src = meta.module_src.replace("True", "False")
            return meta
    """
    assert callable(hook)
    _import_hooks.append(hook)

#----------------------------------------------------------------------------

def _reconstruct_persistent_obj(meta):
    r"""Hook that is called internally by the `pickle` module to unpickle
    a persistent object.
    """
    meta = EasyDict(meta)
    meta.state = EasyDict(meta.state)
    for hook in _import_hooks:
        meta = hook(meta)
        assert meta is not None

    assert meta.version == _version
    module = _src_to_module(meta.module_src)

    assert meta.type == 'class'
    orig_class = module.__dict__[meta.class_name]
    decorator_class = persistent_class(orig_class)
    obj = decorator_class.__new__(decorator_class)

    setstate = getattr(obj, '__setstate__', None)
    if callable(setstate):
        setstate(meta.state) # pylint: disable=not-callable
    else:
        obj.__dict__.update(meta.state)
    return obj

#----------------------------------------------------------------------------

def _module_to_src(module):
    r"""Query the source code of a given Python module.
    """
    src = _module_to_src_dict.get(module, None)
    if src is None:
        src = inspect.getsource(module)
        _module_to_src_dict[module] = src
        _src_to_module_dict[src] = module
    return src

def _src_to_module(src):
    r"""Get or create a Python module for the given source code.
    """
    module = _src_to_module_dict.get(src, None)
    if module is None:
        module_name = "_imported_module_" + uuid.uuid4().hex
        module = types.ModuleType(module_name)
        sys.modules[module_name] = module
        _module_to_src_dict[module] = src
        _src_to_module_dict[src] = module
        exec(src, module.__dict__) # pylint: disable=exec-used
    return module

#----------------------------------------------------------------------------

def _check_pickleable(obj):
    r"""Check that the given object is pickleable, raising an exception if
    it is not. This function is expected to be considerably more efficient
    than actually pickling the object.
    """
    def recurse(obj):
        if isinstance(obj, (list, tuple, set)):
            return [recurse(x) for x in obj]
        if isinstance(obj, dict):
            return [[recurse(x), recurse(y)] for x, y in obj.items()]
        if isinstance(obj, (str, int, float, bool, bytes, bytearray)):
            return None # Python primitive types are pickleable.
        if f'{type(obj).__module__}.{type(obj).__name__}' in ['numpy.ndarray', 'torch.Tensor']:
            return None # NumPy arrays and PyTorch tensors are pickleable.
        if is_persistent(obj):
            return None # Persistent objects are pickleable, by virtue of the constructor check.
        return obj
    with io.BytesIO() as f:
        pickle.dump(recurse(obj), f)

#----------------------------------------------------------------------------



from ple.lit_model import *
try:
    from ple.loss import *
except Exception as e:
    print(f'{e} while import loss, consider this command "pip install mmcv-full"')
from ple.base_exp import *
from ple.trainer import *
from ple.datasets import *
import torch, torch.nn as nn, torch.nn.functional as F


# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_lit_model.ipynb.

# %% auto 0
__all__ = ['Scheduler', 'CosineLRScheduler', 'plot_lr_step_schedule', 'fn_schedule_linear_with_warmup',
           'fn_schedule_cosine_with_warmpup_decay_timm', 'get_scheduler', 'print_example', 'LitModel']

# %% ../nbs/01_lit_model.ipynb 4
from loguru import logger
from pytorch_lightning import LightningModule

import torch
import torch.nn as nn
from datetime import datetime, timedelta
from pytorch_lightning.callbacks.progress import TQDMProgressBar
from pytorch_lightning.loggers import TensorBoardLogger
import os
from fastcore.all import *
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import LearningRateMonitor, ModelCheckpoint
try:
    from ple.loss import FocalLoss, BinaryFocalLoss
except Exception as e:
    FocalLoss, BinaryFocalLoss = None, None
import os.path as osp
from torch.optim.lr_scheduler import LambdaLR


# %% ../nbs/01_lit_model.ipynb 6
from typing import Any
class Scheduler:
    """ Parameter Scheduler Base Class
    A scheduler base class that can be used to schedule any optimizer parameter groups.

    Unlike the builtin PyTorch schedulers, this is intended to be consistently called
    * At the END of each epoch, before incrementing the epoch count, to calculate next epoch's value
    * At the END of each optimizer update, after incrementing the update count, to calculate next update's value

    The schedulers built on this should try to remain as stateless as possible (for simplicity).

    This family of schedulers is attempting to avoid the confusion of the meaning of 'last_epoch'
    and -1 values for special behaviour. All epoch and update counts must be tracked in the training
    code and explicitly passed in to the schedulers on the corresponding step or step_update call.

    Based on ideas from:
     * https://github.com/pytorch/fairseq/tree/master/fairseq/optim/lr_scheduler
     * https://github.com/allenai/allennlp/tree/master/allennlp/training/learning_rate_schedulers
    """

    def __init__(self,
                 optimizer: torch.optim.Optimizer,
                 param_group_field: str,
                 noise_range_t=None,
                 noise_type='normal',
                 noise_pct=0.67,
                 noise_std=1.0,
                 noise_seed=None,
                 initialize: bool = True) -> None:
        self.optimizer = optimizer
        self.param_group_field = param_group_field
        self._initial_param_group_field = f"initial_{param_group_field}"
        if initialize:
            for i, group in enumerate(self.optimizer.param_groups):
                if param_group_field not in group:
                    raise KeyError(f"{param_group_field} missing from param_groups[{i}]")
                group.setdefault(self._initial_param_group_field, group[param_group_field])
        else:
            for i, group in enumerate(self.optimizer.param_groups):
                if self._initial_param_group_field not in group:
                    raise KeyError(f"{self._initial_param_group_field} missing from param_groups[{i}]")
        self.base_values = [group[self._initial_param_group_field] for group in self.optimizer.param_groups]
        self.metric = None  # any point to having this for all?
        self.noise_range_t = noise_range_t
        self.noise_pct = noise_pct
        self.noise_type = noise_type
        self.noise_std = noise_std
        self.noise_seed = noise_seed if noise_seed is not None else 42
        self.update_groups(self.base_values)

    def state_dict(self) -> Dict[str, Any]:
        return {key: value for key, value in self.__dict__.items() if key != 'optimizer'}

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        self.__dict__.update(state_dict)

    def get_epoch_values(self, epoch: int):
        return None

    def get_update_values(self, num_updates: int):
        return None

    def step(self, epoch: int, metric: float = None) -> None:
        self.metric = metric
        values = self.get_epoch_values(epoch)
        if values is not None:
            values = self._add_noise(values, epoch)
            self.update_groups(values)

    def step_update(self, num_updates: int, metric: float = None):
        self.metric = metric
        values = self.get_update_values(num_updates)
        if values is not None:
            values = self._add_noise(values, num_updates)
            self.update_groups(values)

    def update_groups(self, values):
        if not isinstance(values, (list, tuple)):
            values = [values] * len(self.optimizer.param_groups)
        for param_group, value in zip(self.optimizer.param_groups, values):
            if 'lr_scale' in param_group:
                param_group[self.param_group_field] = value * param_group['lr_scale']
            else:
                param_group[self.param_group_field] = value

    def _add_noise(self, lrs, t):
        if self._is_apply_noise(t):
            noise = self._calculate_noise(t)
            lrs = [v + v * noise for v in lrs]
        return lrs

    def _is_apply_noise(self, t) -> bool:
        """Return True if scheduler in noise range."""
        apply_noise = False
        if self.noise_range_t is not None:
            if isinstance(self.noise_range_t, (list, tuple)):
                apply_noise = self.noise_range_t[0] <= t < self.noise_range_t[1]
            else:
                apply_noise = t >= self.noise_range_t
        return apply_noise

    def _calculate_noise(self, t) -> float:
        g = torch.Generator()
        g.manual_seed(self.noise_seed + t)
        if self.noise_type == 'normal':
            while True:
                # resample if noise out of percent limit, brute force but shouldn't spin much
                noise = torch.randn(1, generator=g).item()
                if abs(noise) < self.noise_pct:
                    return noise
        else:
            noise = 2 * (torch.rand(1, generator=g).item() - 0.5) * self.noise_pct
        return noise

# %% ../nbs/01_lit_model.ipynb 7
class CosineLRScheduler(Scheduler):
    """
    Cosine decay with restarts.
    This is described in the paper https://arxiv.org/abs/1608.03983.

    Inspiration from
    https://github.com/allenai/allennlp/blob/master/allennlp/training/learning_rate_schedulers/cosine.py

    k-decay option based on `k-decay: A New Method For Learning Rate Schedule` - https://arxiv.org/abs/2004.05909
    """

    def __init__(self,
                 optimizer: torch.optim.Optimizer,
                 t_initial: int,
                 lr_min: float = 0.,
                 cycle_mul: float = 1.,
                 cycle_decay: float = 1.,
                 cycle_limit: int = 1,
                 warmup_t=0,
                 warmup_lr_init=0,
                 warmup_prefix=False,
                 t_in_epochs=True,
                 noise_range_t=None,
                 noise_pct=0.67,
                 noise_std=1.0,
                 noise_seed=42,
                 k_decay=1.0,
                 initialize=True) -> None:
        super().__init__(
            optimizer, param_group_field="lr",
            noise_range_t=noise_range_t, noise_pct=noise_pct, noise_std=noise_std, noise_seed=noise_seed,
            initialize=initialize)

        assert t_initial > 0
        assert lr_min >= 0
        if t_initial == 1 and cycle_mul == 1 and cycle_decay == 1:
            _logger.warning("Cosine annealing scheduler will have no effect on the learning "
                           "rate since t_initial = t_mul = eta_mul = 1.")
        self.t_initial = t_initial
        self.lr_min = lr_min
        self.cycle_mul = cycle_mul
        self.cycle_decay = cycle_decay
        self.cycle_limit = cycle_limit
        self.warmup_t = warmup_t
        self.warmup_lr_init = warmup_lr_init
        self.warmup_prefix = warmup_prefix
        self.t_in_epochs = t_in_epochs
        self.k_decay = k_decay
        if self.warmup_t:
            self.warmup_steps = [(v - warmup_lr_init) / self.warmup_t for v in self.base_values]
            super().update_groups(self.warmup_lr_init)
        else:
            self.warmup_steps = [1 for _ in self.base_values]

    def _get_lr(self, t):
        if t < self.warmup_t:
            lrs = [self.warmup_lr_init + t * s for s in self.warmup_steps]
        else:
            if self.warmup_prefix:
                t = t - self.warmup_t

            if self.cycle_mul != 1:
                i = math.floor(math.log(1 - t / self.t_initial * (1 - self.cycle_mul), self.cycle_mul))
                t_i = self.cycle_mul ** i * self.t_initial
                t_curr = t - (1 - self.cycle_mul ** i) / (1 - self.cycle_mul) * self.t_initial
            else:
                i = t // self.t_initial
                t_i = self.t_initial
                t_curr = t - (self.t_initial * i)

            gamma = self.cycle_decay ** i
            lr_max_values = [v * gamma for v in self.base_values]
            k = self.k_decay

            if i < self.cycle_limit:
                lrs = [
                    self.lr_min + 0.5 * (lr_max - self.lr_min) * (1 + math.cos(math.pi * t_curr ** k / t_i ** k))
                    for lr_max in lr_max_values
                ]
            else:
                lrs = [self.lr_min for _ in self.base_values]

        return lrs

    def get_epoch_values(self, epoch: int):
        if self.t_in_epochs:
            return self._get_lr(epoch)
        else:
            return None

    def get_update_values(self, num_updates: int):
        if not self.t_in_epochs:
            return self._get_lr(num_updates)
        else:
            return None

    def get_cycle_length(self, cycles=0):
        cycles = max(1, cycles or self.cycle_limit)
        if self.cycle_mul == 1.0:
            return self.t_initial * cycles
        else:
            return int(math.floor(-self.t_initial * (self.cycle_mul ** cycles - 1) / (1 - self.cycle_mul)))

# %% ../nbs/01_lit_model.ipynb 8
# from timm.scheduler.cosine_lr import CosineLRScheduler, Scheduler


def plot_lr_step_schedule(fn, lr, num_epochs, num_steps_per_epoch):
    import matplotlib.pyplot as plt
    lrs = []
    steps = num_epochs*num_steps_per_epoch
    for step in range(steps):
        lrs.append(fn(step)*lr)
    # print(f'{min(lrs)=:0.5f}, {max(lrs)=:0.5f}')
    plt.plot(range(steps), lrs)
    plt.show()


def fn_schedule_linear_with_warmup(num_epochs, num_steps_per_epoch,
                                   num_warmup_steps=0, init_lr=0.4, min_lr=0.1):
    num_training_steps = num_epochs*num_steps_per_epoch

    def lr_lambda(current_step: int):
        if current_step < num_warmup_steps:
            x = (1-init_lr)*(current_step / num_warmup_steps)+init_lr
            return x

        total_step = (num_training_steps-num_warmup_steps)
        current_step = current_step-num_warmup_steps
        rt = min_lr+(1-min_lr)*(1-current_step/total_step)
        return rt
    return lr_lambda


def fn_schedule_cosine_with_warmpup_decay_timm(num_epochs, num_steps_per_epoch, num_epochs_per_cycle,
                                               num_warmup_epochs=1, init_lr=0.4, min_lr=0.1, cycle_decay=.8, interval='step'):
    lr = 1
    num_cycles = num_epochs // num_epochs_per_cycle
    optim = torch.optim.SGD(nn.Linear(1, 1).parameters(), lr)
    m = 1 if interval == 'epoch' else num_steps_per_epoch
    logger.info(f'num_cycles={num_cycles}')
    schedule = CosineLRScheduler(optim,
                                 t_initial=num_epochs_per_cycle*m,
                                 lr_min=min_lr*lr,
                                 cycle_decay=cycle_decay,
                                 cycle_limit=num_cycles, warmup_t=num_warmup_epochs*m, warmup_lr_init=init_lr*lr,
                                 )

    def get_lr(step): return schedule._get_lr(step)[0]
    return get_lr


def get_scheduler(optimizer, lr_schedule_fn, interval='step', verbose=False):
    scheduler = {
        "scheduler": LambdaLR(optimizer, lr_schedule_fn, -1, verbose=verbose),
        "interval": interval,
        "frequency": 1,
    }
    return scheduler


# %% ../nbs/01_lit_model.ipynb 11
def print_example():
    str = """
#---- HYPER
EPOCHS = 30
LR = 1e-4
BZ = 8
GPUS = 1
NUM_WORKERS = 4            
STRATEGY = 'dp'
# --- INIT dataset
train_ds = #
val_ds = #

dl_train = torch.utils.data.DataLoader(train_ds, BZ, num_workers=NUM_WORKERS, shuffle=True)
dl_val = torch.utils.data.DataLoader(train_ds, BZ, num_workers=NUM_WORKERS, shuffle=False)
# ---- Lr scheduler
sched = fn_schedule_cosine_with_warmpup_decay_timm(
    num_epochs=EPOCHS,
    num_steps_per_epoch=len(dl_train),
    num_epochs_per_cycle=EPOCHS//2,
    min_lr=1/100,
    cycle_decay=0.7,
)
# --- Optimizer
optim = lambda params:torch.optim.Adam(params, lr=LR)

# ---- LitModel
class CustomLit(LitModel):
    def forward(self, batch):
        # Get loss

    def training_step(self, batch, idx):
        out = self.forward(batch)
        return out['loss']

    def validation_step(self, batch, idx):
        out = self.forward(batch)
        self.log('val/loss', out['loss'], prog_bar=True, on_epoch=True)

lit = CustomLit(model,create_optimizer_fn=optim,
                               create_lr_scheduler_fn=sched, loss_fn=nn.CrossEntropyLoss())    

#---------------- Train

trainer = get_trainer('trans_lit', EPOCHS, gpus=GPUS, overfit_batches=0.05,
                     monitor={'metric': 'val/loss', 'mode': 'min'}, strategy=STRATEGY)
trainer.fit(trans_lit, dl_train, dl_val)
    """
    print(str)


# %% ../nbs/01_lit_model.ipynb 13
class LitModel(LightningModule):
    def __init__(self, model,
                 create_optimizer_fn=None,
                 create_lr_scheduler_fn=None,
                 loss_fn=nn.CrossEntropyLoss()
                ):

        super().__init__()
        store_attr()

    def configure_optimizers(self):
        """
            Setup optimizer and scheduler
        """
        assert self.create_optimizer_fn is not None

        optimizer = self.create_optimizer_fn(self.model.parameters())
        scheduler = get_scheduler(optimizer, self.create_lr_scheduler_fn)
        return [optimizer], [scheduler]

    def forward(self, x):
        return self.model(x)

    def validation_step(self, batch, batch_idx):
        x, y = batch[:2]
        logits = self(x)
        loss = self.loss_fn(logits, y)
        preds = logits.softmax(1).argmax(1)
        accs = (y == preds).float().mean()
        self.log("val_loss", loss, rank_zero_only=True, prog_bar=True,
                 on_step=False, on_epoch=True)
        self.log("val_acc", accs, rank_zero_only=True, prog_bar=True,
                 on_step=False, on_epoch=True)
        return loss

    def training_step(self, batch, batch_idx):
        x, y = batch[:2]
        logits = self(x)
        loss = self.loss_fn(logits, y)

        preds = logits.softmax(1).argmax(1)
        accs = (y == preds).float().mean()

        self.log("training_loss", loss, prog_bar=True,
                 rank_zero_only=True, on_epoch=True)
        self.log("training_accuracy", accs, prog_bar=True,
                 rank_zero_only=True, on_epoch=True)
        return loss




# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/ple/',
                'doc_host': 'https://anhvth.github.io',
                'git_url': 'https://github.com/anhvth/pl_exp/',
                'lib_path': 'ple'},
  'syms': { 'ple.all': {},
            'ple.base_exp': { 'ple.base_exp.BaseExp': ('base_exp.html#baseexp', 'ple/base_exp.py'),
                              'ple.base_exp.BaseExp.__init__': ('base_exp.html#baseexp.__init__', 'ple/base_exp.py'),
                              'ple.base_exp.BaseExp.__repr__': ('base_exp.html#baseexp.__repr__', 'ple/base_exp.py'),
                              'ple.base_exp.BaseExp.get_data_loader': ('base_exp.html#baseexp.get_data_loader', 'ple/base_exp.py'),
                              'ple.base_exp.BaseExp.get_lr_scheduler': ('base_exp.html#baseexp.get_lr_scheduler', 'ple/base_exp.py'),
                              'ple.base_exp.BaseExp.get_model': ('base_exp.html#baseexp.get_model', 'ple/base_exp.py'),
                              'ple.base_exp.BaseExp.get_optimizer': ('base_exp.html#baseexp.get_optimizer', 'ple/base_exp.py'),
                              'ple.base_exp.BaseExp.get_trainer': ('base_exp.html#baseexp.get_trainer', 'ple/base_exp.py'),
                              'ple.base_exp.BaseExp.merge': ('base_exp.html#baseexp.merge', 'ple/base_exp.py'),
                              'ple.base_exp.BaseExp.plot_lr_sche': ('base_exp.html#baseexp.plot_lr_sche', 'ple/base_exp.py')},
            'ple.datasets': { 'ple.datasets.VideoFrameDataset': ('datasets.html#videoframedataset', 'ple/datasets.py'),
                              'ple.datasets.VideoFrameDataset.__getitem__': ( 'datasets.html#videoframedataset.__getitem__',
                                                                              'ple/datasets.py'),
                              'ple.datasets.VideoFrameDataset.__init__': ('datasets.html#videoframedataset.__init__', 'ple/datasets.py'),
                              'ple.datasets.VideoFrameDataset.__len__': ('datasets.html#videoframedataset.__len__', 'ple/datasets.py'),
                              'ple.datasets.VideoFrameDataset.check_if_extracted': ( 'datasets.html#videoframedataset.check_if_extracted',
                                                                                     'ple/datasets.py'),
                              'ple.datasets.VideoFrameDataset.get_images': ( 'datasets.html#videoframedataset.get_images',
                                                                             'ple/datasets.py')},
            'ple.lit_model': { 'ple.lit_model.CosineLRScheduler': ('lit_model.html#cosinelrscheduler', 'ple/lit_model.py'),
                               'ple.lit_model.CosineLRScheduler.__init__': ( 'lit_model.html#cosinelrscheduler.__init__',
                                                                             'ple/lit_model.py'),
                               'ple.lit_model.CosineLRScheduler._get_lr': ('lit_model.html#cosinelrscheduler._get_lr', 'ple/lit_model.py'),
                               'ple.lit_model.CosineLRScheduler.get_cycle_length': ( 'lit_model.html#cosinelrscheduler.get_cycle_length',
                                                                                     'ple/lit_model.py'),
                               'ple.lit_model.CosineLRScheduler.get_epoch_values': ( 'lit_model.html#cosinelrscheduler.get_epoch_values',
                                                                                     'ple/lit_model.py'),
                               'ple.lit_model.CosineLRScheduler.get_update_values': ( 'lit_model.html#cosinelrscheduler.get_update_values',
                                                                                      'ple/lit_model.py'),
                               'ple.lit_model.LitModel': ('lit_model.html#litmodel', 'ple/lit_model.py'),
                               'ple.lit_model.LitModel.__init__': ('lit_model.html#litmodel.__init__', 'ple/lit_model.py'),
                               'ple.lit_model.LitModel.configure_optimizers': ( 'lit_model.html#litmodel.configure_optimizers',
                                                                                'ple/lit_model.py'),
                               'ple.lit_model.LitModel.forward': ('lit_model.html#litmodel.forward', 'ple/lit_model.py'),
                               'ple.lit_model.LitModel.training_step': ('lit_model.html#litmodel.training_step', 'ple/lit_model.py'),
                               'ple.lit_model.LitModel.validation_step': ('lit_model.html#litmodel.validation_step', 'ple/lit_model.py'),
                               'ple.lit_model.Scheduler': ('lit_model.html#scheduler', 'ple/lit_model.py'),
                               'ple.lit_model.Scheduler.__init__': ('lit_model.html#scheduler.__init__', 'ple/lit_model.py'),
                               'ple.lit_model.Scheduler._add_noise': ('lit_model.html#scheduler._add_noise', 'ple/lit_model.py'),
                               'ple.lit_model.Scheduler._calculate_noise': ( 'lit_model.html#scheduler._calculate_noise',
                                                                             'ple/lit_model.py'),
                               'ple.lit_model.Scheduler._is_apply_noise': ('lit_model.html#scheduler._is_apply_noise', 'ple/lit_model.py'),
                               'ple.lit_model.Scheduler.get_epoch_values': ( 'lit_model.html#scheduler.get_epoch_values',
                                                                             'ple/lit_model.py'),
                               'ple.lit_model.Scheduler.get_update_values': ( 'lit_model.html#scheduler.get_update_values',
                                                                              'ple/lit_model.py'),
                               'ple.lit_model.Scheduler.load_state_dict': ('lit_model.html#scheduler.load_state_dict', 'ple/lit_model.py'),
                               'ple.lit_model.Scheduler.state_dict': ('lit_model.html#scheduler.state_dict', 'ple/lit_model.py'),
                               'ple.lit_model.Scheduler.step': ('lit_model.html#scheduler.step', 'ple/lit_model.py'),
                               'ple.lit_model.Scheduler.step_update': ('lit_model.html#scheduler.step_update', 'ple/lit_model.py'),
                               'ple.lit_model.Scheduler.update_groups': ('lit_model.html#scheduler.update_groups', 'ple/lit_model.py'),
                               'ple.lit_model.fn_schedule_cosine_with_warmpup_decay_timm': ( 'lit_model.html#fn_schedule_cosine_with_warmpup_decay_timm',
                                                                                             'ple/lit_model.py'),
                               'ple.lit_model.fn_schedule_linear_with_warmup': ( 'lit_model.html#fn_schedule_linear_with_warmup',
                                                                                 'ple/lit_model.py'),
                               'ple.lit_model.get_scheduler': ('lit_model.html#get_scheduler', 'ple/lit_model.py'),
                               'ple.lit_model.plot_lr_step_schedule': ('lit_model.html#plot_lr_step_schedule', 'ple/lit_model.py'),
                               'ple.lit_model.print_example': ('lit_model.html#print_example', 'ple/lit_model.py')},
            'ple.loss': { 'ple.loss.BinaryFocalLoss': ('loss.html#binaryfocalloss', 'ple/loss.py'),
                          'ple.loss.BinaryFocalLoss.__init__': ('loss.html#binaryfocalloss.__init__', 'ple/loss.py'),
                          'ple.loss.BinaryFocalLoss.forward': ('loss.html#binaryfocalloss.forward', 'ple/loss.py'),
                          'ple.loss.FocalLoss': ('loss.html#focalloss', 'ple/loss.py'),
                          'ple.loss.FocalLoss.__init__': ('loss.html#focalloss.__init__', 'ple/loss.py'),
                          'ple.loss.FocalLoss.forward': ('loss.html#focalloss.forward', 'ple/loss.py'),
                          'ple.loss.py_sigmoid_focal_loss': ('loss.html#py_sigmoid_focal_loss', 'ple/loss.py'),
                          'ple.loss.reduce_loss': ('loss.html#reduce_loss', 'ple/loss.py'),
                          'ple.loss.sigmoid_focal_loss': ('loss.html#sigmoid_focal_loss', 'ple/loss.py'),
                          'ple.loss.weight_reduce_loss': ('loss.html#weight_reduce_loss', 'ple/loss.py')},
            'ple.persistance': { 'ple.persistance.EasyDict': ('persistance.html#easydict', 'ple/persistance.py'),
                                 'ple.persistance.EasyDict.__delattr__': ('persistance.html#easydict.__delattr__', 'ple/persistance.py'),
                                 'ple.persistance.EasyDict.__getattr__': ('persistance.html#easydict.__getattr__', 'ple/persistance.py'),
                                 'ple.persistance.EasyDict.__setattr__': ('persistance.html#easydict.__setattr__', 'ple/persistance.py'),
                                 'ple.persistance.Logger': ('persistance.html#logger', 'ple/persistance.py'),
                                 'ple.persistance.Logger.__enter__': ('persistance.html#logger.__enter__', 'ple/persistance.py'),
                                 'ple.persistance.Logger.__exit__': ('persistance.html#logger.__exit__', 'ple/persistance.py'),
                                 'ple.persistance.Logger.__init__': ('persistance.html#logger.__init__', 'ple/persistance.py'),
                                 'ple.persistance.Logger.close': ('persistance.html#logger.close', 'ple/persistance.py'),
                                 'ple.persistance.Logger.flush': ('persistance.html#logger.flush', 'ple/persistance.py'),
                                 'ple.persistance.Logger.write': ('persistance.html#logger.write', 'ple/persistance.py'),
                                 'ple.persistance._check_pickleable': ('persistance.html#_check_pickleable', 'ple/persistance.py'),
                                 'ple.persistance._module_to_src': ('persistance.html#_module_to_src', 'ple/persistance.py'),
                                 'ple.persistance._reconstruct_persistent_obj': ( 'persistance.html#_reconstruct_persistent_obj',
                                                                                  'ple/persistance.py'),
                                 'ple.persistance._src_to_module': ('persistance.html#_src_to_module', 'ple/persistance.py'),
                                 'ple.persistance.ask_yes_no': ('persistance.html#ask_yes_no', 'ple/persistance.py'),
                                 'ple.persistance.call_func_by_name': ('persistance.html#call_func_by_name', 'ple/persistance.py'),
                                 'ple.persistance.construct_class_by_name': ( 'persistance.html#construct_class_by_name',
                                                                              'ple/persistance.py'),
                                 'ple.persistance.copy_files_and_create_dirs': ( 'persistance.html#copy_files_and_create_dirs',
                                                                                 'ple/persistance.py'),
                                 'ple.persistance.format_time': ('persistance.html#format_time', 'ple/persistance.py'),
                                 'ple.persistance.get_dtype_and_ctype': ('persistance.html#get_dtype_and_ctype', 'ple/persistance.py'),
                                 'ple.persistance.get_module_dir_by_obj_name': ( 'persistance.html#get_module_dir_by_obj_name',
                                                                                 'ple/persistance.py'),
                                 'ple.persistance.get_module_from_obj_name': ( 'persistance.html#get_module_from_obj_name',
                                                                               'ple/persistance.py'),
                                 'ple.persistance.get_obj_by_name': ('persistance.html#get_obj_by_name', 'ple/persistance.py'),
                                 'ple.persistance.get_obj_from_module': ('persistance.html#get_obj_from_module', 'ple/persistance.py'),
                                 'ple.persistance.get_top_level_function_name': ( 'persistance.html#get_top_level_function_name',
                                                                                  'ple/persistance.py'),
                                 'ple.persistance.import_hook': ('persistance.html#import_hook', 'ple/persistance.py'),
                                 'ple.persistance.is_persistent': ('persistance.html#is_persistent', 'ple/persistance.py'),
                                 'ple.persistance.is_pickleable': ('persistance.html#is_pickleable', 'ple/persistance.py'),
                                 'ple.persistance.is_top_level_function': ('persistance.html#is_top_level_function', 'ple/persistance.py'),
                                 'ple.persistance.is_url': ('persistance.html#is_url', 'ple/persistance.py'),
                                 'ple.persistance.list_dir_recursively_with_ignore': ( 'persistance.html#list_dir_recursively_with_ignore',
                                                                                       'ple/persistance.py'),
                                 'ple.persistance.make_cache_dir_path': ('persistance.html#make_cache_dir_path', 'ple/persistance.py'),
                                 'ple.persistance.open_url': ('persistance.html#open_url', 'ple/persistance.py'),
                                 'ple.persistance.persistent_class': ('persistance.html#persistent_class', 'ple/persistance.py'),
                                 'ple.persistance.set_cache_dir': ('persistance.html#set_cache_dir', 'ple/persistance.py'),
                                 'ple.persistance.tuple_product': ('persistance.html#tuple_product', 'ple/persistance.py')},
            'ple.trainer': { 'ple.trainer.get_exp_by_file': ('trainer.html#get_exp_by_file', 'ple/trainer.py'),
                             'ple.trainer.get_rank': ('trainer.html#get_rank', 'ple/trainer.py'),
                             'ple.trainer.get_trainer': ('trainer.html#get_trainer', 'ple/trainer.py'),
                             'ple.trainer.is_interactive': ('trainer.html#is_interactive', 'ple/trainer.py'),
                             'ple.trainer.train': ('trainer.html#train', 'ple/trainer.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_trainer.ipynb.

# %% auto 0
__all__ = ['is_interactive', 'get_trainer', 'get_rank', 'get_exp_by_file', 'train']

# %% ../nbs/05_trainer.ipynb 4
from pytorch_lightning.callbacks import LearningRateMonitor, ModelCheckpoint
from pytorch_lightning.callbacks.progress import TQDMProgressBar
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning import Trainer
import os.path as osp
from loguru import logger

def is_interactive():
    try:
        shell = get_ipython().__class__.__name__
        return True
    except:
        return False


def get_trainer(exp_name=None, 
                max_epochs=None, 
                gpus=1,
                monitor=dict(metric="val/loss", mode="min"), 
                save_every_n_epochs=1, 
                save_top_k=1, 
                use_version=True, 
                strategy=None,
                refresh_rate=5, 
                **kwargs):
    callbacks = []
    plt_logger = None
    if exp_name is not None:
        rld = osp.join("lightning_logs", exp_name)
        cur_num_exps = len(os.listdir(rld)) if osp.exists(rld) else 0
        version = f"{cur_num_exps:02d}"
        root_log_dir = osp.join("lightning_logs", exp_name, version)
        if monitor is None:
            filename = None
        else:
            metric = monitor["metric"].replace('/', '_')
            filename = "{epoch}_{"+metric+":0.4f}"
        

        # logger.info(f'Tensorboard: tensorboard --logdir {root_log_dir}')

        callback_ckpt = ModelCheckpoint(
            dirpath=osp.join(root_log_dir, "ckpts"),
            monitor=monitor['metric'] if monitor is not None else None, 
            mode=monitor['mode'] if monitor is not None else 'min',
            filename=filename,
            save_last=True,
            every_n_epochs=save_every_n_epochs,
            save_top_k=save_top_k,
        )
        plt_logger = TensorBoardLogger(
            osp.join(root_log_dir, "tb_logs"),
        )
        callbacks.append(callback_ckpt)

    callbacks.append(TQDMProgressBar(refresh_rate=5))
    callbacks.append(LearningRateMonitor(logging_interval="step"))
    

    if strategy is None:
        if is_interactive():
            logger.info("gpus={}, Interactive mode, force strategy=dp", gpus)
            strategy = 'dp'
        elif gpus < 2:
            logger.info("gpus={}, , force strategy=dp", gpus)
            strategy = 'dp'
        else:
            strategy="ddp"
            logger.info("gpus={}, strategy=ddp, set strategy='dp' if this", gpus)
            
    trainer = Trainer(
        accelerator ='gpu',
        devices=gpus,
        max_epochs=max_epochs,
        strategy=strategy,
        callbacks=callbacks,
        log_every_n_steps=10,
        logger=plt_logger, **kwargs
    )
    return trainer

# %% ../nbs/05_trainer.ipynb 8
import os
import os.path as osp
# # from argparse import ArgumentParser
# from pytorch_lightning import Trainer, seed_everything
# # from ple.all import get_trainer, BaseExp
import shutil
from fastcore.script import *
from fastcore.utils import *
from loguru import logger

def get_rank() -> int:
    import torch.distributed as dist
    if not dist.is_available():
        return 0
    if not dist.is_initialized():
        return 0
    return dist.get_rank()

def get_exp_by_file(exp_file):
    """
        Params:
        exp_file: Path to exp
        
    """
    try:
        import importlib
        import os
        import sys
        sys.path.append(os.path.dirname(exp_file))
        # import ipdb; ipdb.set_trace()
        current_exp = importlib.import_module(
            os.path.basename(exp_file).split(".")[0])
        current_exp = importlib.reload(current_exp)
        exp = current_exp.Exp()
        return exp
    except Exception:
        raise ImportError(
            "{} doesn't contains class named 'Exp'".format(exp_file))



# %% ../nbs/05_trainer.ipynb 10
@call_parse
def train(
    cfg_path:Param('Path to config'),
    devices: Param('GPUS indices', default=1, type=int),
    accelerator: Param('cpu or gpu', default='gpu', type=str),
    opts: Param('Additional configs', default='', type=str, required=False),
):
    cfg = get_exp_by_file(cfg_path)
    if len(opts):
        cfg.merge(opts.replace('=',' ').split(' '))
    cfg.devices = devices
    
    data = cfg.get_data_loader()

    model = cfg.get_model(create_lr_scheduler_fn=cfg.get_lr_scheduler(), 
            create_optimizer_fn=cfg.get_optimizer())
    trainer = cfg.get_trainer(devices)
    try:
        trainer.fit(model, data)
    except Exception as e:
        import traceback
        traceback.print_exc()
    finally:
        if get_rank() == 0:
            out_path = osp.join(trainer.log_dir, osp.basename(cfg_path))
            logger.info('cp {} {}', cfg_path, out_path)
            shutil.copy(cfg_path, out_path)




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/06_datasets.ipynb.

# %% auto 0
__all__ = ['VideoFrameDataset']

# %% ../nbs/06_datasets.ipynb 2
import torch.utils.data as td
from avcv.all import *
from fastcore.all import *

class VideoFrameDataset(td.Dataset):
    def __init__(self, path_to_video, img_preproc=None, out_dir=None, img_size=(416, 416), fps=10, ext='png'):
        if out_dir is None:
            vname = get_name(path_to_video)
            out_dir = f'/tmp/ple/VideoFrameDataset/extracted_video_for_inference/{vname}_fps_{fps}'
#         self.out_dir = out_dir
        store_attr('img_preproc, out_dir, img_size, fps, ext')
    
        self.video = mmcv.VideoReader(path_to_video)
        self.paths = self.get_images(path_to_video)
        
    def __len__(self):
        return len(self.paths)
    
    def __getitem__(self, idx):
        img = mmcv.imread(self.paths[idx])
        if self.img_preproc is not None:
            x = self.img_preproc(img)
            return dict(x=x, img=img)
        return dict(img=img)
    
    def get_images(self, path_to_video):
        if not self.check_if_extracted(path_to_video):
            _im = mmcv.imrescale(self.video[0], self.img_size)
            im_h, im_w = _im.shape[:2]
            mmcv.mkdir_or_exist(self.out_dir)
            cmd = f"ffmpeg -n  -i {path_to_video} -s {im_w}x{im_h} -vf fps={self.fps} {self.out_dir}/%06d.{self.ext}"
            print(cmd)
            os.system(cmd)
#         else:
#             print('Skip extracting using ffmpeg since files are already extracted')
        paths = glob(f'{self.out_dir}/*.{self.ext}')
        return list(sorted(paths))
    
    def check_if_extracted(self, path_to_video):
        exist_num = len(glob(f'{self.out_dir}/*.{self.ext}'))
        expected_num = len(self.video)*self.fps/self.video.fps
        exist_rate = exist_num/expected_num
        if exist_rate<0.9:
            print(f'{exist_rate=}')
        return exist_rate>0.9





# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_base_exp.ipynb.

# %% auto 0
__all__ = ['BaseExp']

# %% ../nbs/04_base_exp.ipynb 2
import ast
import pprint
from abc import ABCMeta, abstractmethod
from typing import Dict
from tabulate import tabulate

import torch
from torch.nn import Module
import pytorch_lightning as pl
from torch.optim.lr_scheduler import _LRScheduler

import os.path as osp
from fastcore.all import *


class BaseExp(metaclass=ABCMeta):
    """
        import torch.utils.data as td
        import pytorch_lightning as pl

        class PLData(pl.LightningDataModule):
            def __init__(self, **kwargs):
                super().__init__()
                store_attr()

            def train_dataloader(self):
                # dataset = 
                return td.DataLoader(dataset, self.batch_size, num_workers=self.num_workers)

            def val_dataloader(self):
                # dataset = 
                return td.DataLoader(dataset, self.batch_size, num_workers=self.num_workers)

        class TriStageExp(BaseExp):

            def __init__(self, exp_name='EXPNAME', 
                         batch_size=64, 
                         num_workers=2, 
                         devices=2,
                         strategy='dp', 
                         **kwargs):
                super().__init__()
                store_attr(**kwargs)

            def get_model(self):
                dl = self.get_data_loader().train_dataloader()
                sched = fn_schedule_cosine_with_warmpup_decay_timm(
                    num_epochs=self.max_epochs,
                    num_steps_per_epoch=len(dl)//self.devices,
                    num_epochs_per_cycle=self.max_epochs//self.num_lr_cycles,
                    min_lr=1/100,
                    cycle_decay=0.7,
                )
                optim = lambda params:torch.optim.Adam(params)

                return MyLit(self.model, create_optimizer_fn=optim,
                                           create_lr_scheduler_fn=sched)

            def get_data_loader(self):
                return PLData(self.batch_size, num_workers=self.num_workers)

            def get_trainer(self, **kwargs):
                from ple.trainer import get_trainer
                return get_trainer(self.exp_name, 
                                      max_epochs=self.max_epochs, 
                                      gpus=self.devices,
                                   strategy=self.strategy,
                                   **kwargs,

                                  )
            exp = TriStageExp( batch_size=10, exp_name='hi', devices=2)
            print(exp)
    """

    def __init__(self):
        # All hyper params should be listed here
        self.accelerator = 'gpu'
        self.max_epochs:int = 100
        self.devices:int = 1
        
        self.schedule_type = 'cosine'
        self.num_lr_cycles = 3
        self.min_lr = 1/100
        self.cycle_decay = 0.7
        
    @abstractmethod
    def get_model(self) -> Module:
        pass

    @abstractmethod
    def get_data_loader(
        self,
    ) -> pl.LightningDataModule:
        pass

    def get_optimizer(self) -> torch.optim.Optimizer:
        """
            Examples:
                return lambda params:torch.optim.Adam(params)
        """
        pass


    def get_lr_scheduler(self, train_loader_len=None):
        if train_loader_len is None:
            data = self.get_data_loader()
            data.setup(None)
            #-------
            num_epochs = self.max_epochs
            train_loader = data.train_dataloader()
            train_loader_len = len(train_loader)
        
        if self.schedule_type == 'cosine':
            from ple.lit_model import fn_schedule_cosine_with_warmpup_decay_timm
            create_schedule_fn = fn_schedule_cosine_with_warmpup_decay_timm(
                num_epochs=self.max_epochs,
                num_steps_per_epoch=train_loader_len//self.devices,
                num_epochs_per_cycle=self.max_epochs//self.num_lr_cycles,
                min_lr=self.min_lr,
                cycle_decay=self.cycle_decay,
            )
        elif self.schedule_type == 'linear':
            from ple.lit_model import fn_schedule_linear_with_warmup
            create_schedule_fn = fn_schedule_linear_with_warmup(
                num_epochs=self.trainer.max_epochs,
                num_steps_per_epoch=train_loader_len//self.devices
            )
            
        else:
            raise NotImplementedError
        return create_schedule_fn

        
    def __repr__(self):
        table_header = ["keys", "values"]
        exp_table = [
            (str(k), pprint.pformat(v))
            for k, v in vars(self).items()
            if not k.startswith("_")
        ]
        return tabulate(exp_table, headers=table_header, tablefmt="fancy_grid")
    
    def merge(self, cfg_list):
        assert len(cfg_list) % 2 == 0
        for k, v in zip(cfg_list[0::2], cfg_list[1::2]):
            # only update value with same key
            if hasattr(self, k):
                src_value = getattr(self, k)
                src_type = type(src_value)
                if src_value is not None and src_type != type(v):
                    try:
                        v = src_type(v)
                    except Exception:
                        v = ast.literal_eval(v)
                print(f'Set {k}={v}')
                setattr(self, k, v)
        
    def get_trainer(self):
        from ple.trainer import get_trainer
        return get_trainer(self.exp_name, 
                              max_epochs=self.max_epochs, 
                              gpus=self.devices,
                              trainer_kwargs=dict(
                                  accelerator=self.accelerator,
                              ))
    def plot_lr_sche(self):
        data = self.get_data_loader()
        step_per_epoch = len(data.train_dataloader())
        sched = self.get_lr_scheduler(step_per_epoch)
        plot_lr_step_schedule(sched, self.lr, self.epochs, step_per_epoch)
